// hyperapp-types

// TypeScript Version: 2.8
// Extra type definitions for hyperapp V1

/** Action implements type */
export type ActionImplements<Self, State> = {[key in keyof Self]: ActionImplementItem<Self, State, key>};

type ActionImplementItem<ActionImpls, State, PropName extends keyof ActionImpls> = (
    ActionImplementFunctionLazy<State>
    |
    ActionImplementFunction<State>
    |
    NestedActionImplements<ActionImpls, State, PropName>
);

type ActionImplementFunctionLazy<State> = (data?: any) => (state: State, actions: object) => HyperappActionResult<State>;
type ActionImplementFunction<State> = (data?: any) => HyperappActionResult<State>;

type NestedActionImplements<ActionImpls, State, PropName extends keyof ActionImpls> = (
    PropName extends ((keyof State) & (keyof ActionImpls)) ? ActionImplements<ActionImpls[PropName], State[PropName]> : never
);

/** Wired actions type (generated by type of action implements) */
export type WiredActions<State, ActImpls extends ActionImplements<ActImpls, State>> = WiredActionsInner<State, ActImpls>;

type WiredActionsInner<State, ActImpls> = {[key in keyof ActImpls]: WiredActionItem<State, ActImpls, key>};

type WiredActionItem<State, ActImpls, PropName extends (keyof ActImpls)> = (
    ActImpls[PropName] extends ((data?: infer P) => (state: State, actions: infer A2) => (infer R)) ? ((data?: P) => R) :
    ActImpls[PropName] extends ((data?: infer P) => (infer R))                                      ? ((data?: P) => R) :
    // nested actions
    PropName extends (keyof State) ? WiredActionsInner<State[PropName], ActImpls[PropName]> :
    // invalid
    'ERROR: invalid action type'
);

/** extract parameter type from action */
export type ParamType<A> = (
    A extends ((data: infer P) => any) ? P     :
    never
);

/**
 * hyperapp ActionResult
 *  (copy of hyperapp. If it depends on hyperapp, hyperapp will be installed doubly at npm and an
 * error will occur with duplication of type definition.)
 */
type HyperappActionResult<State> = Partial<State> | Promise<any> | null | void;